@@ -27,7 +27,11 @@
 #include <QCryptographicHash>
 #include <QDomElement>
 #include <QStringList>
+#if QT_VERSION >= 0x050000
+#include <QUrlQuery>
+#else
 #include <QUrl>
+#endif
 
 #include "QXmppSasl_p.h"
 #include "QXmppUtils.h"
@@ -90,7 +94,7 @@ void QXmppSaslAuth::setValue(const QByteArray &value)
 void QXmppSaslAuth::parse(const QDomElement &element)
 {
     m_mechanism = element.attribute("mechanism");
-    m_value = QByteArray::fromBase64(element.text().toAscii());
+    m_value = QByteArray::fromBase64(element.text().toLatin1());
 }
 
 void QXmppSaslAuth::toXml(QXmlStreamWriter *writer) const
@@ -120,7 +124,7 @@ void QXmppSaslChallenge::setValue(const QByteArray &value)
 
 void QXmppSaslChallenge::parse(const QDomElement &element)
 {
-    m_value = QByteArray::fromBase64(element.text().toAscii());
+    m_value = QByteArray::fromBase64(element.text().toLatin1());
 }
 
 void QXmppSaslChallenge::toXml(QXmlStreamWriter *writer) const
@@ -178,7 +182,7 @@ void QXmppSaslResponse::setValue(const QByteArray &value)
 
 void QXmppSaslResponse::parse(const QDomElement &element)
 {
-    m_value = QByteArray::fromBase64(element.text().toAscii());
+    m_value = QByteArray::fromBase64(element.text().toLatin1());
 }
 
 void QXmppSaslResponse::toXml(QXmlStreamWriter *writer) const
@@ -433,15 +437,23 @@ bool QXmppSaslClientFacebook::respond(const QByteArray &challenge, QByteArray &r
         return true;
     } else if (m_step == 1) {
         // parse request
+#if QT_VERSION >= 0x050000
+        QUrlQuery requestUrl(challenge);
+#else
         QUrl requestUrl;
         requestUrl.setEncodedQuery(challenge);
+#endif
         if (!requestUrl.hasQueryItem("method") || !requestUrl.hasQueryItem("nonce")) {
             warning("QXmppSaslClientFacebook : Invalid challenge, nonce or method missing");
             return false;
         }
 
         // build response
+#if QT_VERSION >= 0x050000
+        QUrlQuery responseUrl;
+#else
         QUrl responseUrl;
+#endif
         responseUrl.addQueryItem("access_token", password());
         responseUrl.addQueryItem("api_key", username());
         responseUrl.addQueryItem("call_id", 0);
@@ -449,7 +461,11 @@ bool QXmppSaslClientFacebook::respond(const QByteArray &challenge, QByteArray &r
         responseUrl.addQueryItem("nonce", requestUrl.queryItemValue("nonce"));
         responseUrl.addQueryItem("v", "1.0");
 
+#if QT_VERSION >= 0x050000
+        response = responseUrl.query().toUtf8();
+#else
         response = responseUrl.encodedQuery();
+#endif
         m_step++;
         return true;
     } else {
